"""
File   : halfedges.py

Project: MakeHuman info@makehuman.org, http://www.makehuman.org/

App    : makehuman

Authors : Manuel Bastioni

A plugin to improve the basic data structure, adding halfedges.
Prefix for new properties added to original classes: *he_*

.. image:: ../images/hedge_struct.png

New properties added to base classes:

- Object3D.he_hedges
- Object3D.he_hedgeCalculated
- Vert.he_hedge
- Face.he_hedge

Require:

- base modules

(c) MakeHuman team 2001-2008
This file is part of MakeHuman.

MakeHuman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
MakeHuman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with MakeHuman.  If not, see <http://www.gnu.org/licenses/>.

.. include:: includes/code_guidelines.txt
"""

__docformat__ = 'restructuredtext'


class HEdge:
    """
    A halfedge data structure (abbreviated as HalfedgeDS,
    or HDS for template parameters) is an edge-centered data structure capable
    of maintaining incidence information of vertices, edges and faces,
    for example for planar maps, polyhedra, or other orientable, two-dimensional
    surfaces embedded in arbitrary dimension.

    .. image:: ../images/hedge_struct.png
       :alt: edge structure diagram

    For more details, see cgafaq_ , or flipcode_.   

    .. _cgafaq: http://www.cgafaq.info/wiki/Half_edge_implementation

    .. _flipcode: http://www.flipcode.com/archives/The_Half-Edge_Data_Structure.shtml

    """
    def __init__(self,heId,heNextId,heTwinId,heVertId,heFaceId):
        """
        :Parameter heId: string, key-identifier of this half edge
        :Parameter heNextId: string, key-identifier of next hedge
        :Parameter heTwinId: string, key-identifier of twin hedge
        :Parameter heVertId: integer, index of first vert of this hedge
        :Parameter heFaceId: integer, index of face pointed by this hedge


        .. image:: ../images/hedge_verts.png
            :alt: verts used to init half edge

        Initialize two types of attributes: keys and references.

        Keys
        -----
        Keys are IDs to be used in hedges dictionary.

        - self.twinID: the ID to find the twin of this hedge.
        - self.nextID: the ID to find the next of this hedge.
        - self.ID: the ID to find this hedge.
        - self.fID: the ID to find the face of this hedge.
        - self.vID: the ID to find the vert of this hedge.

        We use keys and dictionary to quick identify the adiacency
        elements without use any search loop, but only  using the
        geometric property illustrated below (we assume a coerent mesh):

        .. image:: ../images/hedge_id.png
           :alt: geometric way to find ID of adiacency elements

        .. image:: ../images/hedge_twin.png

        .. image:: ../images/hedge_next.png

        References
        ----------
        In a successive step (Object3D.linkHEdges function),
        we store directly the references
        of the adiacent elements, instead their ID. This permit a more
        quick access. These references attributes (inital value = None)
        are:

        - *self.twin*: reference to twin hedge
        - *self.next*: reference to netx hedge
        - *self.face*: reference to face
        - *self.vert*: reference to vert
        - *self.sub_vert*: index of subdivided vert in subdivided object
         

        """
        self.twinID = heTwinId
        self.nextID = heNextId
        self.ID = heId
        self.fID = heFaceId
        self.vID = heVertId
        self.twin = None
        self.next = None
        self.face = None
        self.vert = None
        
    

    def __str__(self):
        """
        If an hEdge obj is called with *print* method,
        it's returned his ID, that's  good way to represent in human readable
        string the edge, because indicating the start and end verts.
        """
        return self.ID



def addHedge(ob,heId1,heNextId1,heTwinId1,vIndex,fIndex ):
    

    he = HEdge(heId1,heNextId1,heTwinId1,vIndex,fIndex)
    #print vIndex   
    v = ob.verts[vIndex]
    f = ob.faces[fIndex]        
    f.he_hedge = he #New property for base face class
    v.he_hedge = he #New property for base vert class
    he.vert = v
    he.face = f
    ob.he_hedges[heId1] = he


def linkHEdges(ob):
        """
        Make the *linked list*, storing the reference of various hedge
        objects into each hEdge. Before this linking, we use their keys
        to retrieve them in hedges dictionary, but this is slow.

        In example, to get the twin of an hEdge, before the linking,
        we need to get it from hedges dictionary, using *nextID* as
        key:

        ::

          nextID = he.getNextID()
          heNext = self.hEdges[nextID]

        instead, after linking, it is just:

        ::

          he.setNext(heNext)

        we can access **directly** to twin edge,
        simply writing:

        ::

          he.twin

        Unfortunately it's impossible to assign directly
        references. Infact, we need to use a **two loop technique**:

        - A loop to create all hedges objects, and store them in
          hedges dictionary.
        - A loop to link them as references.

        So, when all hEdges are into edges
        dictionary, we use *linkHEdges()* to retrieve
        and assign them as reference.
        """
        print "linking hedges", len(ob.he_hedges)
        for he in ob.he_hedges.values():            
            try:                
                he.next = ob.he_hedges[he.nextID]           
            except:
                continue

        for he in ob.he_hedges.values():
            try:                
                he.twin = ob.he_hedges[he.twinID]                
            except:
                continue



def calcHedgeData(ob):
        """
        Calculate the *half edges* data
        structure, adding hedges and improving faces

        .. image:: ../images/face.png

        - heId = dataList[1]+"-"+dataList[2], a string key to store
          the half edge 1 in hedges dictionary.
        - heNextId = dataList[2]+"-"+dataList[3], a string key to find
          the "next" half edge 1 in hedges dictionary.
        - heNextId = dataList[2]+"-"+dataList[1], a string key to find
          the "twin" half edge 1 in hedges dictionary.

        In example, looking the image above:

        Hedge ["vo-v1"] has as next the hedge["v1-v2"] and as twin the
        hedge["v1-v0"].

        """
        ob.he_hedges = {} #New property added to base obj
        fIndex = 0        
        
        for f in ob.faces:
            
            vIndex0 = str(f.verts[0].idx)
            vIndex1 = str(f.verts[1].idx)
            vIndex2 = str(f.verts[2].idx) 

            heId1 = vIndex0+"-"+vIndex1
            heNextId1 = vIndex1+"-"+vIndex2
            heTwinId1 = vIndex1+"-"+vIndex0

            heId2 = vIndex1+"-"+vIndex2
            heNextId2 = vIndex2+"-"+vIndex0
            heTwinId2 = vIndex2+"-"+vIndex1

            heId3 = vIndex2+"-"+vIndex0
            heNextId3 = vIndex0+"-"+vIndex1
            heTwinId3 = vIndex0+"-"+vIndex2

            vIndex0 = f.verts[0].idx
            vIndex1 = f.verts[1].idx
            vIndex2 = f.verts[2].idx
            
            addHedge(ob,heId1,heNextId1,heTwinId1,vIndex0,fIndex)
            addHedge(ob,heId2,heNextId2,heTwinId2,vIndex1,fIndex)
            addHedge(ob,heId3,heNextId3,heTwinId3,vIndex2,fIndex)
            
            fIndex += 1
            
        linkHEdges(ob)
        ob.he_hedgeCalculated = 1

def edgesSharedByVert(v):
        i = 0
        try:    
            startHEdge = v.he_hedge
        except:
            print v.idx
        edgesShared = [startHEdge]       
        
        try:
            loopHedge = startHEdge.twin.next
            while loopHedge != startHEdge:                
                edgesShared.append(loopHedge)
                if  i == 10:                
                    break
                i += 1
                try:
                    loopHedge = loopHedge.twin.next
                except:
                    break
                       
        except:            
            if startHEdge.next.next.twin:
                loopHedge = startHEdge.next.next.twin
                while loopHedge != startHEdge:                    
                    edgesShared.append(loopHedge)
                    if  i == 2:                
                        break
                    i += 1
                    if loopHedge.next.next.twin:
                        loopHedge = loopHedge.next.next.twin
                    else:
                        break
            else:                
                edgesShared.append(startHEdge.next)                
        return edgesShared


def hedgesSharedByFace(f):
    hEdgesShared = []
    startHEdge = f.he_hedge
    hEdgesShared.append(startHEdge)
    loopHedge = startHEdge.next
    x = 0
    if loopHedge:
        while loopHedge != startHEdge and x < 2:
            hEdgesShared.append(loopHedge)
            loopHedge = loopHedge.next
            x += 1
    else:
        return []
    return hEdgesShared

def facesSharedByVert(v):
    
    hEdges = edgesSharedByVert(v)
    facesShared = []
    for he in hEdges:
        facesShared.append(he.face)

    return facesShared

#def subvertIdxSharedByVert(v):
    
    #hEdges = edgesSharedByVert(v)
    #subvertsIdxs = []
    #for he in hEdges:
        #subvertsIdxs.append(he.subVertID) #TODO: getter

    #return subvertsIdxs

    
def vertsSharedbyVert(v):
    vertsShared = []
    facesShared = facesSharedByVert(v)
    for f in facesShared:
        for ve in f.verts:
            if ve != v and ve not in vertsShared:
                vertsShared.append(ve)
    return vertsShared
