"""
File   : subdivision.py

Project: MakeHuman info@makehuman.org, http://www.makehuman.org/

App    : makehuman

Authors : Manuel Bastioni

A plugin to subdivide the mesh.

Require:

- base modules

(c) MakeHuman team 2001-2008
This file is part of MakeHuman.

MakeHuman is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
MakeHuman is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with MakeHuman.  If not, see <http://www.gnu.org/licenses/>.

.. include:: includes/code_guidelines.txt
"""

__docformat__ = 'restructuredtext'

import time
import module3d


def newSubVert1(v, objIdx):
    loopVertices = v.vertsShared()
    vSubd = module3d.Vert(v.co[0],v.co[1],v.co[2],v.idx,objIdx)
    vSubd.sub_type = 1
    n = len(loopVertices)

    if n == 3:
        b = 0.1875
        a = 1-(n*b)
    elif n > 3:
        b = 3.0/(8*n)
        a = 1.0-(n*b)
    else:
        a = 0.75 #3/4
        b = 0.125 #1/8
    vSubd.sub_data = [v,a,b,loopVertices]

    calcSubCoord1(vSubd)
    return vSubd

def calcSubCoord1(vSubd):
    #Calculate the position for existing original verts

    v = vSubd.sub_data[0]
    a = vSubd.sub_data[1]
    b = vSubd.sub_data[2]
    loopVertices = vSubd.sub_data[3]
    subdX = a*v.co[0]
    subdY = a*v.co[1]
    subdZ = a*v.co[2]
    noX = a*v.no[0]
    noY = a*v.no[1]
    noZ = a*v.no[2]
    texU = a*v.uv[0]
    texV = a*v.uv[1]
    for lv in loopVertices:
        subdX += b*lv.co[0]
        subdY += b*lv.co[1]
        subdZ += b*lv.co[2]
        noX += b*lv.no[0]
        noY += b*lv.no[1]
        noZ += b*lv.no[2]
        texU += b*lv.uv[0]
        texV += b*lv.uv[1]

    vSubd.co = [subdX,subdY,subdZ]
    vSubd.no = [noX,noY,noZ]
    vSubd.uv = [texU,texV]
    vSubd.color = v.color

def calcSubCoord2(vSubd):
    #Calculate the position for add edge vert type 2

    v1 = vSubd.sub_data[0]
    v2 = vSubd.sub_data[1]
    v3 = vSubd.sub_data[2]
    v4 = vSubd.sub_data[3]
    subdX = 0.375*(v1.co[0] + v2.co[0]) + 0.125*(v3.co[0] + v4.co[0])
    subdY = 0.375*(v1.co[1] + v2.co[1]) + 0.125*(v3.co[1] + v4.co[1])
    subdZ = 0.375*(v1.co[2] + v2.co[2]) + 0.125*(v3.co[2] + v4.co[2])
    noX = 0.375*(v1.no[0] + v2.no[0]) + 0.125*(v3.no[0] + v4.no[0])
    noY = 0.375*(v1.no[1] + v2.no[1]) + 0.125*(v3.no[1] + v4.no[1])
    noZ = 0.375*(v1.no[2] + v2.no[2]) + 0.125*(v3.no[2] + v4.no[2])
    texU = 0.375*(v1.uv[0] + v2.uv[0]) + 0.125*(v3.uv[0] + v4.uv[0])
    texV = 0.375*(v1.uv[1] + v2.uv[1]) + 0.125*(v3.uv[1] + v4.uv[1])
    r = (v1.color[0] + v2.color[0])/2
    g = (v1.color[1] + v2.color[1])/2
    b = (v1.color[2] + v2.color[2])/2
    a = (v1.color[3] + v2.color[3])/2
    vSubd.co = [subdX,subdY,subdZ]
    vSubd.no = [noX,noY,noZ]
    vSubd.uv = [texU,texV]
    vSubd.color = [r,g,b,a]

def calcSubCoord3(vSubd):
    #Calculate the position for add edge vert type 3
    v1 = vSubd.sub_data[0]
    v2 = vSubd.sub_data[1]

    subdX = 0.5*(v1.co[0] + v2.co[0])
    subdY = 0.5*(v1.co[1] + v2.co[1])
    subdZ = 0.5*(v1.co[2] + v2.co[2])
    noX = 0.5*(v1.no[0] + v2.no[0])
    noY = 0.5*(v1.no[1] + v2.no[1])
    noZ = 0.5*(v1.no[2] + v2.no[2])
    texU = 0.5*(v1.uv[0] + v2.uv[0])
    texV = 0.5*(v1.uv[1] + v2.uv[1])
    r = (v1.color[0] + v2.color[0])/2
    g = (v1.color[1] + v2.color[1])/2
    b = (v1.color[2] + v2.color[2])/2
    a = (v1.color[3] + v2.color[3])/2
    vSubd.co = [subdX,subdY,subdZ]
    vSubd.no = [noX,noY,noZ]
    vSubd.uv = [texU,texV]
    vSubd.color = [r,g,b,a]


def subdivide(ob, scene):

    t = time.time()
    subdividedObj = scene.getObject(ob.name+".sub")
    if subdividedObj:
        print "Sub data present"
        for v in subdividedObj.verts:
            if v.sub_type == 1:
                calcSubCoord1(v)
            if v.sub_type == 2:
                calcSubCoord2(v)
            if v.sub_type == 3:
                calcSubCoord3(v)
        for v in subdividedObj.verts:
            v.update()

    else:
        print "Calculating sub data"
        calcSubData(ob, scene)
    print "time for subdivision: %s"%(time.time()-t)

def addEdgeVert(v0, v1, v2, face, edges, objIdx):

    if v0.idx > v1.idx:
        key = str(v0.idx)+"-"+str(v1.idx)
        edge = [v0,v1]
    else:
        key = str(v1.idx)+"-"+str(v0.idx)
        edge = [v1,v0]

    if not edges.has_key(key):
        vSubd = module3d.Vert(0,0,0,-1,objIdx)
        vSubd.sub_data = edge + [v2]
        vSubd.sub_type = 3
        edges[key] = vSubd
    else:
        vSubd = edges[key]
        vSubd.sub_type = 2
        vSubd.sub_data.append(v2)

    face.sub_verts.append(vSubd)



def calcSubData(ob, scene):

    t1 = time.time()
    subObjName = ob.name+".sub"

    t3 = time.time()
    subdividedObj = scene.newObj(subObjName)
    for v in ob.verts:
        vSubd = newSubVert1(v,subdividedObj.idx)
        subdividedObj.verts.append(vSubd)
        subdividedObj.colors.append([v.color[0],v.color[1], v.color[2], v.color[3]])

    print "time to create new obj and map verts: ",  time.time()-t3

    t4 = time.time()
    edgeVerts = {}
    for f in ob.faces:
        f.sub_verts = []
        v0 = f.verts[0]
        v1 = f.verts[1]
        v2 = f.verts[2]
        addEdgeVert(v0, v1, v2, f, edgeVerts, subdividedObj.idx)
        addEdgeVert(v1, v2, v0, f, edgeVerts, subdividedObj.idx)
        addEdgeVert(v2, v0, v1, f, edgeVerts, subdividedObj.idx)

    for ev in edgeVerts.values():
            if ev.sub_type == 3:
                calcSubCoord3(ev)
            if ev.sub_type == 2:
                calcSubCoord2(ev)
            ev.idx = len(subdividedObj.verts)
            subdividedObj.verts.append(ev)
            subdividedObj.colors.append([ev.color[0],ev.color[1], ev.color[2], ev.color[3]])
    print "time to add edge verts: ", time.time()-t4

    fg = module3d.FaceGroup("subobj")
    fg.parentName = subObjName
    subdividedObj.facesGroups.append(fg)
    for f in ob.faces:
        gID = f.groupName
        v0 = subdividedObj.verts[f.verts[0].idx]
        v2 = subdividedObj.verts[f.verts[1].idx]
        v4 = subdividedObj.verts[f.verts[2].idx]
        v1 = f.sub_verts[0]
        v3 = f.sub_verts[1]
        v5 = f.sub_verts[2]
        f1 = module3d.Face(v0,v1,v5)
        f2 = module3d.Face(v1,v2,v3)
        f3 = module3d.Face(v1,v3,v5)
        f4 = module3d.Face(v5,v3,v4)
        f1.groupName = gID
        f2.groupName = gID
        f3.groupName = gID
        f4.groupName = gID
        fg.faces.append(f1)
        fg.faces.append(f2)
        fg.faces.append(f3)
        fg.faces.append(f4)
        subdividedObj.faces.append(f1)
        subdividedObj.faces.append(f2)
        subdividedObj.faces.append(f3)
        subdividedObj.faces.append(f4)

    scene.assignSelectionID(subdividedObj)
    if ob.texture:
        subdividedObj.texture = ob.texture
        pass


    scene.update()
    print "Time to calc and send subdata: ", time.time()-t1
